from typing import Callable, List, Tuple
from unittest import TestCase

from ..drivers.memory import MemoryDriver
from ..tree import branch, _BranchDriver
from ..scheduler import SimpleScheduler


class RecursiveTest(TestCase):
    def _oneRecursiveCall(
        self, scaleFactor: float
    ) -> List[Tuple[float, float]]:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        recursive, scheduler2 = branch(scheduler1, scaleFactor)
        calls = []
        scheduler2.callAt(
            1.0,
            lambda: calls.append((scheduler1.now(), scheduler2.now())),
        )
        driver.advance()
        return calls

    def test_scaling(self) -> None:
        calls = self._oneRecursiveCall(1.0)
        self.assertEqual(calls, [(1.0, 1.0)])
        calls = self._oneRecursiveCall(3.0)
        self.assertEqual(calls, [(1 / 3.0, 1.0)])
        calls = self._oneRecursiveCall(1 / 3.0)
        self.assertEqual(calls, [(3.0, 1.0)])

    def test_changeScaling(self) -> None:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        recursive, scheduler2 = branch(scheduler1, 2.0)
        calls = []
        scheduler2.callAt(
            1.0,
            lambda: calls.append((scheduler1.now(), scheduler2.now())),
        )
        driver.advance(1 / 4)
        recursive.scaleFactor *= 2.0
        self.assertEqual(driver.advance(), 1 / 8)
        self.assertEqual(calls, [((1 / 4) + (1 / 8), 1.0)])

    def test_unscheduleNoOp(self) -> None:
        """
        Unscheduling when not scheduled is a no-op.
        """
        _BranchDriver(SimpleScheduler(MemoryDriver())).unschedule()

    def test_unpausePauseUnpause(self) -> None:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        recursive, scheduler2 = branch(scheduler1, scaleFactor=2)
        recursive.pause()
        self.assertEqual(scheduler2.now(), 0.0)
        driver.advance(500)
        self.assertEqual(scheduler2.now(), 0.0)
        recursive.unpause()
        self.assertEqual(scheduler2.now(), 0.0)
        driver.advance(10)
        self.assertEqual(scheduler2.now(), 20.0)
        recursive.unpause()
        self.assertEqual(scheduler2.now(), 20.0)
        driver.advance(10)
        self.assertEqual(scheduler2.now(), 40.0)

    def test_moveSooner(self) -> None:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        recursive, scheduler2 = branch(scheduler1)
        calls: list[tuple[float, float]] = []
        recursive.unpause()

        recordTimestamp = timestampRecorder(calls, scheduler1, scheduler2)

        scheduler2.callAt(1.0, recordTimestamp)
        scheduler2.callAt(0.5, recordTimestamp)
        driver.advance(0.6)
        self.assertEqual(calls, [(0.6, 0.6)])

    def test_pausing(self) -> None:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        recursive, scheduler2 = branch(scheduler1)
        calls = []
        scheduler2.callAt(
            1.0,
            lambda: calls.append((scheduler1.now(), scheduler2.now())),
        )
        scheduler2.callAt(
            2.0,
            lambda: calls.append((scheduler1.now(), scheduler2.now())),
        )
        self.assertEqual(calls, [])
        driver.advance(1.5)
        self.assertEqual(calls, [(1.5, 1.5)])
        calls[:] = []
        recursive.pause()
        # paused at 1.5, with 0.5 left until second call (at 2.0)
        driver.advance(2.7)
        # move to 4.2, still 0.5 left, no call yet
        self.assertEqual(calls, [])
        self.assertEqual(2.7 + 1.5, driver.now())
        self.assertEqual(1.5, scheduler2.now())
        recursive.unpause()
        driver.advance(0.5)
        self.assertEqual(2.7 + 1.5 + 0.5, driver.now())
        self.assertEqual(1.5 + 0.5, scheduler2.now())
        self.assertEqual(calls, [(2.7 + 1.5 + 0.5, 2.0)])

    def test_doubleUnpause(self) -> None:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        scaleFactor = 2.0
        recursive, scheduler2 = branch(scheduler1, scaleFactor)
        recursive.pause()
        baseTime = 1000.0
        driver.advance(baseTime)
        calls = []
        localDelta = 5.0
        scaledDelta = localDelta / scaleFactor
        scheduler2.callAt(
            scheduler2.now() + localDelta,
            lambda: calls.append((scheduler1.now(), scheduler2.now())),
        )
        recursive.unpause()
        driver.advance(1.0)
        self.assertEqual(calls, [])
        recursive.unpause()
        driver.advance(1.0)
        self.assertEqual(calls, [])
        recursive.unpause()
        driver.advance(0.5)
        self.assertEqual(calls, [(baseTime + scaledDelta, localDelta)])

    def test_idling(self) -> None:
        scheduler1 = SimpleScheduler(driver := MemoryDriver())
        recursive, scheduler2 = branch(scheduler1)
        calls: list[tuple[float, float]] = []
        recordTimestamp = timestampRecorder(calls, scheduler1, scheduler2)
        onlyCall = scheduler2.callAt(1.0, recordTimestamp)
        self.assertTrue(driver.isScheduled())
        onlyCall.cancel()
        self.assertFalse(driver.isScheduled())


def timestampRecorder(
    calls: list[tuple[float, float]],
    scheduler1: SimpleScheduler,
    scheduler2: SimpleScheduler,
) -> Callable[[], None]:
    def recorder() -> None:
        calls.append((scheduler1.now(), scheduler2.now()))

    return recorder



# -*- test-case-name: fritter.test.test_json -*-
"""
Serialize L{Scheduler} and the calls scheduled with it to JSON.

To use this module:

    - create a L{JSONRegistry} at module scope

    - decorate your functions and methods with L{JSONRegistry.function} and
      L{JSONRegistry.method}

    - implement L{JSONable.typeCodeForJSON}, L{JSONable.asJSON},
      L{JSONableInstance.fromJSON} as appropriate, until C{mypy} passes on your
      code

    - instantiate a L{JSONableScheduler} with an appropriate driver

    - schedule your functions and methods using it

    - save it with L{JSONRegistry.save}

    - load it later with L{JSONRegistry.load}
"""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import timedelta
from typing import Any, Callable, Generic, ParamSpec, Protocol, Type, TypeVar
from zoneinfo import ZoneInfo

from datetype import DateTime, fromisoformat
from fritter.repeat import EveryDelta

from ..boundaries import Cancellable, RepeatingWork, TimeDriver
from ..drivers.datetime import DateTimeDriver
from ..repeat import RecurrenceRule, Repeater
from ..scheduler import Scheduler

LoadContext = TypeVar("LoadContext", contravariant=True)
"""
Each L{JSONRegistry} has its own special object passed along to
L{JSONableInstance.fromJSON}, which can be any object.  This TypeVar describes
its type.
"""
LoadContextInv = TypeVar("LoadContextInv")
"""
Like L{LoadContext} but invariant.
"""
JSONObject = dict[str, Any]
"""
A loose description of a JSON-dumpable object.
"""
# TODO: tighten JSONObject up to accurately describe valid input to
# C{json.dumps}.


class HasTypeCode(Protocol):
    """
    The L{HasTypeCode} protocol describes an object with a C{typeCodeForJSON}
    method; this is used as a bound on other types such as L{JSONableCallable},
    L{JSONableRepeatable}, and L{JSONableInstance}.
    """

    def typeCodeForJSON(self) -> str:
        """
        Type-code for the serialized JSON object to be looked up later.
        """


class JSONable(HasTypeCode, Protocol):
    """
    Methods that allow a L{JSONRegistry} to serialize an object as JSON.
    """

    def asJSON(self) -> JSONObject:
        """
        Convert this callable to a JSON-serializable dictionary.
        """


class JSONableCallable(JSONable, Protocol):
    """
    A callable that can be serialized as JSON.
    """

    def __call__(self) -> None:
        """
        Do the work of the callable.
        """


JSONableScheduler = Scheduler[DateTime[ZoneInfo], JSONableCallable]
"""
A JSONable scheduler is a L{Scheduler} which tracks time as a L{ZoneInfo}-aware
C{datetype.DateTime} and only accepts L{JSONableCallable} objects as work to be
performed, so that they can be serialized later.

@note: The type-checker can make sure that your callalbles are json-able, but
    they can't make sure that they're registered with the correct
    L{JSONRegistry}; mixing and matching different registries will be a runtime
    error.
"""
# TODO: datetype API docs?


class JSONableRepeatable(JSONable, RepeatingWork, Protocol):
    """
    A callable that can be serialized as JSON, with a signature that can be
    used for repeated calls with L{JSONRegistry.repeatedly}.
    """


class JSONableInstance(JSONable, Protocol[LoadContextInv]):
    """
    A class that conforms to L{JSONableInstance} can be both serialized to and
    deserialized from JSON by a L{JSONRegistry}.  L{JSONableInstance}.
    """

    @classmethod
    def typeCodeForJSON(cls) -> str:
        """
        On a L{JSONableInstance}, the C{typeCodeForJSON} method must be a
        L{classmethod}, because the class needs to be registered for
        I{de}-serialization at class-definition time, before any instances
        exist.

        @see: L{HasTypeCode}
        """

    @classmethod
    def fromJSON(
        cls: Type[JSONableSelf],
        registry: JSONRegistry[LoadContextInv],
        scheduler: JSONableScheduler,
        loadContext: LoadContextInv,
        json: JSONObject,
    ) -> JSONableSelf:
        """
        Load an instance of this type from the given deserialized JSON object,
        which should be in the format returned by an instance of this class's
        L{asJSON <JSONable.asJSON>} method.
        """


JSONableSelf = TypeVar(
    "JSONableSelf",
    bound=JSONableInstance[Any],
    # See comment in JSONRegistry.method about this bound.
)
"""
TypeVar for binding C{self} on methods that want to be serialized by
L{JSONRegistry.method}.
"""

_JSONableType = TypeVar("_JSONableType", bound=HasTypeCode)
"Binding for internal generic tracking of JSONable types."

JSONRepeater = Repeater[
    DateTime[ZoneInfo], JSONableCallable, JSONableRepeatable
]
"A L{Repeater} that is constrained to accept only JSON-serializable types."


def _whatJSON(what: JSONable) -> JSONObject:
    """
    Convert a L{JSONable} into a standard JSON-serializable object format.
    """
    return {
        "type": what.typeCodeForJSON(),
        "data": what.asJSON(),
    }


SomeSignature = ParamSpec("SomeSignature")
"""
A description of the parameters passed to a serializable function.
"""


@dataclass
class SerializableFunction(Generic[SomeSignature]):
    """
    Wrapper around a function that conforms with L{JSONable}.
    """

    original: Callable[SomeSignature, object]
    typeCode: str

    def __call__(
        self, *args: SomeSignature.args, **kwargs: SomeSignature.kwargs
    ) -> None:
        """
        Delegate invocation to the wrapped callable.
        """
        self.original(*args, **kwargs)

    def typeCodeForJSON(self) -> str:
        """
        Return the static type code supplied at construction time.
        """
        return self.typeCode

    def asJSON(self) -> JSONObject:
        """
        Return an empty object.
        """
        return {}


@dataclass
class JSONableBoundMethod(Generic[JSONableSelf]):
    """
    A bound method that conforms to L{JSONable} and can be serialized and
    deserialized.

    @ivar descriptor: the descriptor that this bound method is binding, created
        by L{JSONRegistry.method}.

    @ivar instance: the instance that this bound method is bound to; a
        L{JSONableInstance} of some kind.
    """

    descriptor: JSONableMethodDescriptor[JSONableSelf, Any]
    instance: JSONableSelf

    def __call__(self) -> None:
        """
        Call this bound method's function with its instance as C{self}.
        """
        self.descriptor.func(self.instance)

    def asJSON(self) -> JSONObject:
        """
        Convert this method's instance to a JSON-dumpable dict.
        """
        return self.instance.asJSON()

    def typeCodeForJSON(self) -> str:
        """
        Combine the bound method's name with its owning class's type code.
        """
        return (
            f"{self.instance.typeCodeForJSON()}."
            f"{self.descriptor.func.__name__}"
        )


__JC: Type[JSONableCallable] = JSONableBoundMethod


@dataclass
class JSONableMethodDescriptor(Generic[JSONableSelf, LoadContext]):
    """
    A descriptor that can bind methods into L{JSONableBoundMethod}s.

    @see: L{JSONRegistry.method}

    @ivar registry: The L{JSONRegistry} which created this descriptor.
    @ivar func: The callable underlying method function.
    """

    registry: JSONRegistry[LoadContext]
    func: Callable[[JSONableSelf], None]

    def __set_name__(
        self, cls: Type[JSONableInstance[LoadContext]], name: str
    ) -> None:
        """
        Register the class of the decorated method when the decorator is
        registered with the class, via the protocol of L{object.__set_name__}.
        """
        self.registry._registerJSONableType(cls)

    def __get__(
        self, instance: JSONableSelf, owner: object = None
    ) -> JSONableBoundMethod[JSONableSelf]:
        """
        Bind the decorated method to an instance.
        """
        return JSONableBoundMethod(self, instance)


@dataclass
class JSONableBoundRepeatable(Generic[JSONableSelf]):
    """
    Like L{JSONableBoundMethod}, but for repeating calls.

    @ivar descriptor: the descriptor that this bound method is binding, created
        by L{JSONRegistry.repeatMethod}.

    @ivar instance: the instance that this bound method is bound to; a
        L{JSONableInstance} of some kind.
    """

    descriptor: JSONableRepeatableDescriptor[JSONableSelf, Any]
    instance: JSONableSelf

    def __call__(self, steps: int, stopper: Cancellable) -> None:
        """
        Call this bound method's function with its instance as C{self}.
        """
        self.descriptor.func(self.instance, steps, stopper)

    def asJSON(self) -> JSONObject:
        """
        Convert this method's instance to a JSON-dumpable dict.
        """
        return self.instance.asJSON()

    def typeCodeForJSON(self) -> str:
        """
        Combine the bound method's name with its owning class's type code.
        """
        return (
            f"{self.instance.typeCodeForJSON()}."
            f"{self.descriptor.func.__name__}"
        )


@dataclass
class JSONableRepeatableDescriptor(Generic[JSONableSelf, LoadContext]):
    """
    A descriptor that can bind methods into L{JSONableBoundRepeatable}s.

    @see: L{JSONRegistry.repeatMethod}

    @ivar registry: The L{JSONRegistry} which created this descriptor.
    @ivar func: The callable underlying method function.
    """

    registry: JSONRegistry[LoadContext]
    func: Callable[[JSONableSelf, int, Cancellable], None]

    def __set_name__(
        self, cls: Type[JSONableInstance[LoadContext]], name: str
    ) -> None:
        """
        Register the class of the decorated method when the decorator is
        registered with the class, via the protocol of L{object.__set_name__}.
        """
        self.registry._registerJSONableType(cls)

    def __get__(
        self, instance: JSONableSelf, owner: object = None
    ) -> JSONableBoundRepeatable[JSONableSelf]:
        """
        Bind the decorated method to an instance.
        """
        return JSONableBoundRepeatable(self, instance)


class _JSONableMethodBinder(Protocol[LoadContextInv]):
    """
    Separated description of L{JSONableMethodDescriptor.__get__} that wraps
    L{_JSONableRepeaterWrapper.repeat} to work around U{this issue
    <https://github.com/python/mypy/issues/15822>}.
    """

    def __call__(
        self,
        instance: _JSONableRepeaterWrapper[LoadContextInv],
        owner: object = None,
    ) -> JSONableBoundMethod[_JSONableRepeaterWrapper[LoadContextInv]]:
        """
        Bind the method.
        """


@dataclass
class _SpecificTypeRegistration(Generic[_JSONableType]):
    """
    General container for registering a specific kind of type (instances, plain
    callables, repeated callables).

    @ivar registered: The specific type registration.
    """

    registered: dict[str, _JSONableType] = field(default_factory=dict)

    def add(self, item: _JSONableType) -> _JSONableType:
        """
        Add the given type to the registry, keyed by its C{typeCodeForJSON}.
        """
        self.registered[item.typeCodeForJSON()] = item
        return item

    def get(self, typeCode: str) -> _JSONableType | None:
        """
        Get a type from the registry by the result of its C{typeCodeForJSON}
        classmethod, previously registered by L{_SpecificTypeRegistration.add}.
        """
        return self.registered.get(typeCode, None)


@dataclass
class JSONRegistry(Generic[LoadContext]):
    """
    A L{JSONRegistry} maintains a set of functions and methods (and the classes
    those methods are defined on) that can scheduled against a L{Scheduler} and
    then persisted into a JSON blob to reconstitute that scheduler later.

    Use a L{JSONRegistry} by instantiatint it at module scope and then using
    its methods as decorators on functions and methods.
    """

    # TODO: implement merging multiple registries together so that objects from
    # different libraries can live in the same blob
    _functions: _SpecificTypeRegistration[JSONableCallable] = field(
        default_factory=_SpecificTypeRegistration
    )
    _repeatable: _SpecificTypeRegistration[JSONableRepeatable] = field(
        default_factory=_SpecificTypeRegistration
    )
    _instances: _SpecificTypeRegistration[
        Type[JSONableInstance[LoadContext]]
    ] = field(default_factory=_SpecificTypeRegistration)

    _bindRepeaterConversionMethod: _JSONableMethodBinder[LoadContext] = field(
        init=False
    )

    def _loadOne(
        self,
        json: JSONObject,
        scheduler: JSONableScheduler,
        which: _SpecificTypeRegistration[_JSONableType],
        ctx: LoadContext,
    ) -> _JSONableType:
        """
        Convert the given JSON-dumpable dict into an object of C{_JSONableType}
        via its L{JSONableInstance.fromJSON} classmethod registered in the
        given L{_SpecificTypeRegistration}.
        """
        # inverse of _whatJSON
        typeCode = json["type"]
        blob = json["data"]

        if (it := which.get(typeCode)) is not None:
            return it

        classCode, methodName = typeCode.rsplit(".", 1)
        if (instanceType := self._instances.get(classCode)) is not None:
            # TODO: record allowable instance/method pairs so that we have some
            # confidence that the resulting type here is in fact actually
            # `JSONableType`.  probably the right way to do this is to have
            # _instances live on SpecificTypeRegistration rather than be shared
            # between both repeatable/non-repeatable types
            result: _JSONableType = getattr(
                instanceType.fromJSON(self, scheduler, ctx, blob),
                methodName,
            )
            return result

        # TODO: this is for *long-term* storage of sets of scheduled events, so
        # failing the entire load like this is probably not always an
        # acceptable failure mode.  We should identify any calls that failed to
        # load, remember their uninterpreted blobs to be included again in
        # save() so we don't lose them, then communicate the failure to an
        # object passed to .load()

        raise KeyError(f"cannot interpret type code {repr(typeCode)}")

    def __post_init__(self) -> None:
        """
        Ensure that the glue for serializing repeating calls by converting them
        into regular calls is set up.
        """
        # TODO: something about bound methods of generics (specifically
        # _JSONableRepeaterWrapper[LoadContext].repeat) breaks type
        # inference here so we need an explicit declaration, should report this
        # upstream.
        descriptor: JSONableMethodDescriptor[
            _JSONableRepeaterWrapper[LoadContext], LoadContext
        ] = self.method(_JSONableRepeaterWrapper.repeat)
        # this is happening after the fact, not during a class definition, so
        # we need to take care of the __set_name__ step manually.
        descriptor.__set_name__(
            _JSONableRepeaterWrapper, _JSONableRepeaterWrapper.repeat.__name__
        )
        self._bindRepeaterConversionMethod = descriptor.__get__

    def _repeaterToJSONable(
        self, repeater: JSONRepeater
    ) -> JSONableBoundMethod[_JSONableRepeaterWrapper[LoadContext]]:
        """
        Convert the given L{JSONRepeater} into a method that can be serialized.
        """
        return self._bindRepeaterConversionMethod(
            _JSONableRepeaterWrapper(self, repeater)
        )

    def repeatedly(
        self,
        scheduler: JSONableScheduler,
        rule: RecurrenceRule[DateTime[ZoneInfo]],
        work: JSONableRepeatable,
        reference: DateTime[ZoneInfo] | None = None,
    ) -> None:
        """
        Call C{work} repeatedly, according to C{rule}, with intervals computed
        according to the given C{reference} time (or C{scheduler}'s current
        time, if not given).

        If you want to stop the repetition, C{work} is passed a L{Cancellable}
        that you can call C{.cancel()} on to do so.

        @see: L{fritter.repeat}
        """
        Repeater.new(
            scheduler, rule, work, self._repeaterToJSONable, reference
        ).repeat()

    def function(self, cb: Callable[[], None]) -> JSONableCallable:
        """
        Mark the given 0-argument, None-returning, top-level function as
        possible to serialize within this registry.  It will be serialized by
        its fully-qualified Python name.

        Use like so::

            registry = JSONRegistry[...]()

            @registry.serializedFunction
        """
        return self._functions.add(SerializableFunction(cb, cb.__qualname__))

    def method(
        self, method: Callable[[JSONableSelf], None]
    ) -> JSONableMethodDescriptor[JSONableSelf, LoadContext]:
        """
        Mark the given method, defined at class scope in a class complying with
        the L{JSONableInstance} protocol, as possible to serialize within this
        registry.
        """
        # I want to stipulate that the method I am taking here must have a
        # 'self' parameter whose type is *at least* as strict as
        # JSONableInstance[LoadContext] but may be stricter than that.
        # However, this would require higher-kinded typevars, in order to make
        # JSONableSelf bounded by JSONableInstance[LoadContext] rather than
        # JSONableInstance[Any]: https://github.com/python/typing/issues/548
        return JSONableMethodDescriptor(self, method)

    def repeatFunction(self, cb: RepeatingWork) -> JSONableRepeatable:
        """
        Mark the given function that matches the signature of L{RepeatingWork},
        i.e. one which takes a number of steps and a L{Cancellable} to cancel
        its own repetition, as looked up via its C{__qualname__} attribute.

        @return: a function that mimics the signature of the original function,
            but also conforms to the L{JSONable} protocol.
        """
        return self._repeatable.add(
            SerializableFunction(cb, getattr(cb, "__qualname__"))
        )

    def repeatMethod(
        self, repeatable: Callable[[JSONableSelf, int, Cancellable], None]
    ) -> JSONableRepeatableDescriptor[JSONableSelf, LoadContext]:
        """
        Mark the given method that matches the signature of L{RepeatingWork},
        i.e. one which takes a number of steps and a L{Cancellable} to cancel
        its own repetition, as looked up via its C{__qualname__} attribute.

        @note: The registration occurs when the class is defined, specifically
            in the C{__set_name__} hook in the returned descriptor, so if you
            are calling this in an esoteric context outside a normal class
            definition, be sure to invoke that hook as well.

        @return: a descriptor that mimics the signature of the original method.
        """
        return JSONableRepeatableDescriptor(self, repeatable)

    def _registerJSONableType(
        self, cls: Type[JSONableInstance[LoadContext]]
    ) -> None:
        """
        Mark the given class as serializable via this registry, keyed by its
        L{JSONable.typeCodeForJSON} method.  Used internally as supporting
        L{JSONRegistry.method} and L{JSONRegistry.repeatMethod}.
        """
        self._instances.add(cls)

    def load(
        self,
        runtimeDriver: TimeDriver[float],
        serializedJSON: JSONObject,
        loadContext: LoadContext,
    ) -> JSONableScheduler:
        """
        Load a JSON object in the format serialized from L{JSONRegistry.save}
        and a runtime L{TimeDriver}C{[float]}, returning a
        L{JSONableScheduler}.
        """
        new: JSONableScheduler = Scheduler(
            DateTimeDriver(runtimeDriver),
            counter=int(serializedJSON.get("counter", "0")),
        )
        for callJSON in serializedJSON["scheduledCalls"]:
            when = fromisoformat(callJSON["when"]).replace(
                tzinfo=ZoneInfo(callJSON["tz"])
            )
            what = self._loadOne(
                callJSON["what"], new, self._functions, loadContext
            )
            new.callAt(when, what)
        return new

    def save(self, scheduler: JSONableScheduler) -> JSONObject:
        """
        Serialize the given L{JSONableScheduler} to a single L{JSONObject}
        which can be passed to L{json.dumps} or L{json.loads}.
        """
        # TODO: `self` not used yet here because we're not validating that all
        # the serializable callables here are present in this specific
        # registry, but that would be a good check to have.
        return {
            "scheduledCalls": [
                {
                    "when": item.when.replace(tzinfo=None).isoformat(),
                    "tz": item.when.tzinfo.key,
                    "what": _whatJSON(item.what),
                    "called": item.called,
                    "canceled": item.canceled,
                }
                for item in scheduler.calls()
            ],
            "counter": str(scheduler.counter),
        }


@dataclass
class _JSONableRepeaterWrapper(Generic[LoadContext]):
    """
    Since a L{Scheduler} can only contain C{work} of a given type, which must
    have a 0-argument, C{None}-returning signature, and L{JSONRegistry.save}
    serializes its scheduler by enumerating its future calls, this is a wrapper
    with a special method registered with I{all} L{JSONRegistry}s
    automatically, to provide a JSON-serialization format for a repeated call
    with the repeat-call signature (i.e., C{(steps, stopper)}).

    @see: L{fritter.repeat}

    @ivar jsonRegistry: The specific L{JSONRegistry} to which this
        repeater-wrapper is bound.

    @ivar repeater: The L{JSONable} repeating call that can itself be
        serialized.
    """

    jsonRegistry: JSONRegistry[LoadContext]
    repeater: JSONRepeater

    @classmethod
    def typeCodeForJSON(cls) -> str:
        """
        Return a unique type code (C{"fritter:repetition"}) for this wrapper.
        """
        return "fritter:repetition"

    @classmethod
    def fromJSON(
        cls,
        registry: JSONRegistry[LoadContext],
        scheduler: JSONableScheduler,
        loadContext: LoadContext,
        json: JSONObject,
    ) -> _JSONableRepeaterWrapper[LoadContext]:
        """
        Deserialize a L{_JSONableRepeaterWrapper} from a JSON-dumpable dict
        previously produced by L{_JSONableRepeaterWrapper.asJSON}.
        """
        rule = cls._loadRule(json["rule"])
        what = json["callable"]
        one = registry._loadOne(
            what, scheduler, registry._repeatable, loadContext
        )
        ref = fromisoformat(json["ts"]).replace(tzinfo=ZoneInfo(json["tz"]))
        rep = Repeater(scheduler, rule, one, registry._repeaterToJSONable, ref)
        return cls(registry, rep)

    def _saveRule(self, rule: object) -> object:
        assert isinstance(
            rule, EveryDelta
        ), "Only EveryDelta instances supported so far"
        result: object = rule.delta.__reduce__()[1]
        return result

    @classmethod
    def _loadRule(cls, rule: list[int]) -> EveryDelta:
        return EveryDelta(timedelta(*rule))

    def asJSON(self) -> JSONObject:
        """
        Serialize this L{_JSONableRepeaterWrapper} to a JSON-dumpable dict
        suitable for deserialization with L{_JSONableRepeaterWrapper.fromJSON},
        including its time, IANA timezone identifier, rule function, and
        underlying repeating callable.
        """
        when = self.repeater.reference
        return {
            "ts": when.replace(tzinfo=None).isoformat(),
            "tz": when.tzinfo.key,
            "rule": self._saveRule(self.repeater.rule),
            # TODO: this si just a demo.
            "callable": _whatJSON(self.repeater.work),
            # "convert": is implicitly L{registry._repeaterToJSONable}
            # "scheduler": is what's doing the serializing
        }

    def repeat(self) -> None:
        """
        This C{repeat} method is what is actually serialized, using the
        bound-method support in L{JSONRegistry}.  Internally, it is implicitly
        decorated with each registry's L{JSONRegistry.method} decorator.
        """
        self.repeater.repeat()


__all__ = [
    "JSONableScheduler",
    "JSONObject",
    "JSONRegistry",
]



"""
L{fritter.boundaries} describes the boundaries between different parts of the
system and its interface with your application code.  It contains L{Protocol}s,
L{TypeVar}s, and constant values, but no logic of its own.
"""

from __future__ import annotations
from typing import (
    Any,
    Callable,
    Coroutine,
    Generator,
    Iterator,
    Optional,
    Protocol,
    TypeVar,
)


class PriorityComparable(Protocol):
    """
    Protocol describing an object that can be compared for the purposes of a
    L{PriorityQueue}.
    """

    def __lt__(self, other: Any) -> bool:
        """
        Is C{self} lower priority than C{other}?
        """

    def __le__(self, other: Any) -> bool:
        """
        Is C{self} lower I{or} the same priority as C{other}?
        """


Prioritized = TypeVar("Prioritized", bound=PriorityComparable)
"""
A TypeVar for objects that can be compared for relative priority.
"""

Yield = TypeVar("Yield", covariant=True)
"""
Yield TypeVar for generators and coroutines.
"""
Send = TypeVar("Send", contravariant=True)
"""
Send TypeVar for generators and coroutines.
"""
Return = TypeVar("Return", covariant=True)
"""
Return TypeVar for generators and coroutines.
"""


class CancellableAwaitable(Protocol[Yield, Send, Return]):
    """
    An object which can be both awaited and canceled.
    """

    def __await__(self) -> Generator[Yield, Send, Return]:
        """
        This object may be awaited.
        """

    def cancel(self) -> object:
        """
        This object may be canceled.
        """


AsyncType = TypeVar("AsyncType", bound=CancellableAwaitable[Any, Any, Any])
"TypeVar bound to a L{CancellableAwaitable}"


class PriorityQueue(Protocol[Prioritized]):
    """
    High-level specification of a priority queue.
    """

    def add(self, item: Prioritized) -> None:
        """
        Add an item to the priority queue.
        """

    def get(self) -> Optional[Prioritized]:
        """
        Consume the lowest item from the priority queue.
        """

    def peek(self) -> Optional[Prioritized]:
        """
        Examine the lowest item from the priority queue without modifying the
        queue.
        """

    def remove(self, item: Prioritized) -> bool:
        """
        Remove an item and return whether it was removed or not.
        """

    def __iter__(self) -> Iterator[Prioritized]:
        """
        Iterate all of the values in the priority queue, in an unspecified
        order.
        """


class TimeDriver(Protocol[Prioritized]):
    """
    Driver interface that allows Fritter to schedule objects onto a third party
    library.
    """

    def reschedule(
        self, newTime: Prioritized, work: Callable[[], None]
    ) -> None:
        """
        Schedule C{work} to occur at C{newTime}, removing any previous C{work}
        scheduled by prior calls to C{reschedule}.
        """

    def unschedule(self) -> None:
        """
        Remove any previously-scheduled C{work}.
        """

    def now(self) -> Prioritized:
        """
        Get the current time according to the underlying library.
        """


class Cancellable(Protocol):
    """
    An object that can be cancelled.
    """

    def cancel(self) -> object:
        """
        Cancel an operation in progress.
        """


class RepeatingWork(Protocol):
    """
    L{RepeatingWork} is work that can be performed repeatedly in a loop.
    """

    def __call__(self, steps: int, stopper: Cancellable) -> None:
        """
        Do the work.

        @param steps: The number of steps which have passed since the previous
            invocation.

        @param stopper: An object that you can call cancel() on to stop the
            repetition.
        """


class AsyncDriver(Protocol[AsyncType]):
    """
    An L{AsyncDriver} is an interface to a library that supports awaitables
    (such as the standard library L{asyncio}), to allow L{fritter.repeat.Async}
    to return an awaitable of a library-appropriate type.
    """

    def newWithCancel(self, cancel: Callable[[], None]) -> AsyncType:
        """
        Create a new future-ish object with the given callback to execute when
        it is cancelled.
        """

    def complete(self, asyncObj: AsyncType) -> None:
        """
        The asynchronous operation described by a previous call to this
        L{AsyncDriver}'s L{newWithCancel <AsyncDriver.newWithCancel>} method
        completed successfully; un-suspend any coroutine awaiting it.
        """

    def runAsync(self, coroutine: Coroutine[AsyncType, Any, Any]) -> AsyncType:
        """
        Run the given coroutine which awaits upon L{AsyncType}.

        @note: Whether this starts the given coroutine synchronously or waits
            until the next event-loop tick is implementation-defined.
        """



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
from pprint import pprint
import subprocess

project = "Fritter"
copyright = "2023, Glyph"
author = "Glyph"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.intersphinx",
    "pydoctor.sphinx_ext.build_apidocs",
]

import pathlib

_project_root = pathlib.Path(__file__).parent.parent

# -- Extension configuration ----------------------------------------------
_git_reference = subprocess.run(
    ["git", "rev-parse", "--abbrev-ref", "HEAD"],
    text=True,
    encoding="utf8",
    capture_output=True,
    check=True,
).stdout


print(f"== Environment dump for {_git_reference} ===")
pprint(dict(os.environ))
print("======")


# Try to find URL fragment for the GitHub source page based on current
# branch or tag.

if _git_reference == "HEAD":
    # It looks like the branch has no name.
    # Fallback to commit ID.
    _git_reference = subprocess.getoutput("git rev-parse HEAD")

if os.environ.get("READTHEDOCS", "") == "True":
    rtd_version = os.environ.get("READTHEDOCS_VERSION", "")
    if "." in rtd_version:
        # It looks like we have a tag build.
        _git_reference = rtd_version

_source_root = _project_root / "src"
print("what?")
print(_source_root)
print(_project_root)

pydoctor_args = [
    # pydoctor should not fail the sphinx build, we have another tox environment for that.
    "--intersphinx=https://docs.twisted.org/en/twisted-22.1.0/api/objects.inv",
    "--intersphinx=https://docs.python.org/3/objects.inv",
    "--intersphinx=https://zopeinterface.readthedocs.io/en/latest/objects.inv",
    # TODO: not sure why I have to specify these all twice.

    f"--config={_project_root}/.pydoctor.cfg",
    f"--html-viewsource-base=https://github.com/glyph/fritter/tree/{_git_reference}/src",
    f"--project-base-dir={_source_root}",
    "--html-output={outdir}/api",
    "--privacy=HIDDEN:fritter.test.*",
    "--privacy=HIDDEN:fritter.test",
    "--privacy=HIDDEN:**.__post_init__",
    str(_source_root / "fritter"),
]
pydoctor_url_path = "/en/{rtd_version}/api/"
intersphinx_mapping = {
    "py3": ("https://docs.python.org/3", None),
    "zopeinterface": ("https://zopeinterface.readthedocs.io/en/latest", None),
    "twisted": ("https://docs.twisted.org/en/twisted-22.1.0/api", None),
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]
htmlhelp_basename = "Fritterdoc"



from ..heap import Heap

from unittest import TestCase


class QueueTests(TestCase):
    def setUp(self) -> None:
        self.q: Heap[int] = Heap()

    def test_get(self) -> None:
        self.q.add(3)
        self.q.add(4)
        self.q.add(2)
        self.assertEqual(self.q.get(), 2)
        self.assertEqual(self.q.get(), 3)
        self.assertEqual(self.q.get(), 4)
        self.assertIs(self.q.get(), None)

    def test_peek(self) -> None:
        self.q.add(7)
        self.assertEqual(self.q.peek(), 7)
        self.assertEqual(self.q.peek(), 7)

    def test_remove(self) -> None:
        self.q.add(9)
        self.q.add(10)
        self.q.add(11)
        self.assertEqual(self.q.remove(70), False)
        self.assertEqual(self.q.remove(10), True)
        self.assertEqual(self.q.get(), 9)
        self.assertEqual(self.q.get(), 11)



# -*- test-case-name: fritter.test.test_scheduler -*-
"""
A L{Scheduler} is the core interface of Fritter; a collection of timed calls
scheduled by L{callAt <Scheduler.callAt>} connected to a L{TimeDriver} that
causes them to actually be called.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Callable, Generic, Iterator, TypeVar

from .boundaries import PriorityComparable, PriorityQueue, TimeDriver
from .heap import Heap

WhenT = TypeVar("WhenT", bound=PriorityComparable)
"""
TypeVar for representing a time at which something can occur; a temporal
coordinate in a timekeeping system.
"""
WhatT = TypeVar("WhatT", bound=Callable[[], None])
"""
TypeVar for representing a unit of work that can take place within the context
of a L{Scheduler}.
"""


@dataclass(eq=True, order=True)
class FutureCall(Generic[WhenT, WhatT]):
    """
    A handle to a future call.

    @ivar when: When will this call be run?

    @ivar what: What work will this call perform?

    @ivar id: An ID, unique to the scheduler for identifying this call.
    """

    when: WhenT = field(compare=True)
    what: WhatT = field(compare=False)
    id: int = field(compare=True)
    called: bool = field(compare=False)
    canceled: bool = field(compare=False)
    _canceller: Callable[[FutureCall[WhenT, WhatT]], None] = field(
        compare=False
    )

    def cancel(self) -> None:
        """
        Cancel this L{FutureCall}, making it so that it will not be invoked in
        the future.  If the work described by C{when} has already been called,
        or this call has already been cancelled, do nothing.
        """
        if self.called:
            # nope
            return
        if self.canceled:
            # nope
            return
        self.canceled = True
        self._canceller(self)


@dataclass
class Scheduler(Generic[WhenT, WhatT]):
    """
    A L{Scheduler} allows for scheduling work (of the type C{WhatT}, which must
    be at least a 0-argument None-returning callable) at a given time
    (C{WhenT}, which much be sortable as a L{PriorityComparable}).

    @ivar driver: The L{TimeDriver} that this L{Scheduler} will use.

    @ivar counter: The value for the next ID.
    """

    driver: TimeDriver[WhenT]
    _q: PriorityQueue[FutureCall[WhenT, WhatT]] = field(default_factory=Heap)
    counter: int = 0
    _maxWorkBatch: int = 0xFF

    def __post_init__(self) -> None:
        """
        Ensure that the supplied priority queue is initially empty.
        """
        if self._q.peek() is not None:
            raise ValueError("Priority queue must be initially empty.")

    def now(self) -> WhenT:
        """
        Relay C{now} to our L{TimeDriver}.
        """
        return self.driver.now()

    def calls(self) -> Iterator[FutureCall[WhenT, WhatT]]:
        """
        Iterate through all the L{FutureCall}s previously scheduled by this
        L{Scheduler}'s L{callAt <Scheduler.callAt>} method.
        """
        return iter(self._q)

    def callAt(self, when: WhenT, what: WhatT) -> FutureCall[WhenT, WhatT]:
        """
        Call C{what} at the time C{when} according to the L{TimeDriver}
        associated with this L{Scheduler}.

        @return: a L{FutureCall} that describes the pending call and allows for
            cancelling it.
        """
        self.counter += 1

        def advanceToNow() -> None:
            timestamp = self.driver.now()
            workPerformed = 0
            while (
                (each := self._q.peek()) is not None
                and each.when <= timestamp
                and workPerformed < self._maxWorkBatch
            ):
                popped = self._q.get()
                assert popped is each
                # not sure if there's a more graceful way to put this
                # todo: failure handling
                each.called = True
                each.what()
                workPerformed += 1
            upNext = self._q.peek()
            if upNext is not None:
                self.driver.reschedule(upNext.when, advanceToNow)

        def _cancelCall(toRemove: FutureCall[WhenT, WhatT]) -> None:
            old = self._q.peek()
            self._q.remove(toRemove)
            new = self._q.peek()
            if new is None:
                self.driver.unschedule()
            elif old is None or new is not old:
                self.driver.reschedule(new.when, advanceToNow)

        previously = self._q.peek()
        call = FutureCall(when, what, self.counter, False, False, _cancelCall)
        self._q.add(call)
        currently = self._q.peek()
        # We just added a thing it can't be None even though peek has that
        # signature
        assert currently is not None
        if previously is None or previously.when != currently.when:
            self.driver.reschedule(currently.when, advanceToNow)
        return call


SimpleScheduler = Scheduler[float, Callable[[], None]]
"""
A L{SimpleScheduler} is a L{Scheduler} with a configuration familiar to most
timekeeping systems in Python: time is a L{float} and work to be performed is
any 0-argument callable which returns C{None}.
"""

__all__ = [
    "Scheduler",
    "FutureCall",
    "SimpleScheduler",
    "WhenT",
    "WhatT",
]



from json import dump
from zoneinfo import ZoneInfo

from datetype import DateTime
from datetime import timedelta

from fritter.drivers.datetime import DateTimeDriver
from fritter.drivers.memory import MemoryDriver
from fritter.persistent.json import JSONableScheduler
from fritter.repeat import EveryDelta

from json_instance import MyClass, registry

memoryDriver = MemoryDriver()
scheduler = JSONableScheduler(DateTimeDriver(memoryDriver))
dt = DateTime.now(ZoneInfo(key="America/Los_Angeles"))
memoryDriver.advance(dt.timestamp())
handle = scheduler.callAt(dt + timedelta(seconds=5), MyClass(3).later)
myInstance = MyClass(3)
registry.repeatedly(
    scheduler, EveryDelta(timedelta(seconds=0.5)), myInstance.repeat
)


with open("saved-schedule.json", "w") as f:
    dump(registry.save(scheduler), f)



from __future__ import annotations

from dataclasses import dataclass

from fritter.boundaries import Cancellable
from fritter.persistent.json import JSONableScheduler, JSONObject, JSONRegistry

registry: JSONRegistry[dict[int, MyClass]] = JSONRegistry()


@dataclass
class MyClass:
    value: int

    @classmethod
    def typeCodeForJSON(cls) -> str:
        return ".".join([cls.__module__, cls.__name__])

    def asJSON(self) -> dict[str, object]:
        return {"value": self.value, "id": id(self)}

    @classmethod
    def fromJSON(
        cls,
        registry: JSONRegistry[dict[int, MyClass]],
        scheduler: JSONableScheduler,
        loadContext: dict[int, MyClass],
        json: JSONObject,
    ) -> MyClass:
        loadingID = int(json["id"])
        if loadingID in loadContext:
            return loadContext[loadingID]
        else:
            self = cls(json["value"])
            loadContext[loadingID] = self
            return self

    @registry.method
    def later(self) -> None:
        print("my value is", self.value)

    @registry.repeatMethod
    def repeat(self, steps: int, stopper: Cancellable) -> None:
        print(f"performing {steps} steps at {self.value}")
        self.value += steps
        if self.value > 10:
            stopper.cancel()



from fritter.boundaries import Cancellable
from fritter.repeat import EverySecond, repeatedly
from fritter.drivers.sleep import SleepDriver
from fritter.scheduler import Scheduler
# driver setup
driver = SleepDriver()
start = driver.now()
# repeating work
def work(steps: int, stopper: Cancellable) -> None:
    elapsed = driver.now() - start
    print(f"took {steps} steps at {elapsed:0.2f}")
    if elapsed >= 2.0:
        stopper.cancel()
# kick off scheduler
repeatedly(Scheduler(driver), work, EverySecond(0.05))
steps = driver.block()
print(f"took {steps } steps, then stopped")



from json import dump, load

# from fritter.drivers.memory import MemoryDriver
from fritter.drivers.sleep import SleepDriver

from json_identity import registry

driver = SleepDriver()

with open("saved-id-schedule.json", "r") as f:
    s = registry.load(driver, load(f), {})

driver.block(1.7)

with open("saved-id-schedule.json", "w") as f:
    dump(registry.save(s), f)



from typing import Callable

from datetime import datetime, timedelta
from datetype import DateTime, aware
from zoneinfo import ZoneInfo

from fritter.drivers.memory import MemoryDriver
from fritter.drivers.datetime import DateTimeDriver
from fritter.scheduler import Scheduler

# set up memory driver
memdriver = MemoryDriver()
# set up datetime driver
TZ = ZoneInfo("US/Pacific")
dtdriver = DateTimeDriver(memdriver, TZ)
# set up scheduler
scheduler = Scheduler[DateTime[ZoneInfo], Callable[[], None]](dtdriver)
# create datetime
dt = datetime(2023, 5, 5, tzinfo=TZ)
# advance to the timestamp
memdriver.advance(dt.timestamp())
# define some work
def hi() -> None:
    print("hi", scheduler.now().isoformat())
# schedule the work for 2 days in the future
scheduler.callAt(aware(dt, ZoneInfo) + timedelta(days=2), hi)
memdriver.advance(2*(60 * 60 * 24))



from typing import Callable
from unittest import TestCase

from ..scheduler import FutureCall, SimpleScheduler
from ..drivers.memory import MemoryDriver
from ..heap import Heap


class SchedulerTests(TestCase):
    """
    Tests for L{Scheduler}.
    """

    def test_schedulingSimple(self) -> None:
        """
        Scheduling a call
        """
        driver = MemoryDriver()
        scheduler = SimpleScheduler(driver)
        called = 0

        def callme() -> None:
            nonlocal called
            called += 1

        handle = scheduler.callAt(1.0, callme)
        scheduler.callAt(3.0, callme)
        self.assertEqual(0, called)
        driver.advance(2.0)
        self.assertEqual(1, called)
        self.assertEqual(handle.called, True)
        handle.cancel()  # no-op

    def test_moveSooner(self) -> None:
        driver = MemoryDriver()
        scheduler = SimpleScheduler(driver)
        called = 0

        def callme() -> None:
            nonlocal called
            called += 1

        scheduler.callAt(1.0, callme)
        scheduler.callAt(0.5, callme)
        self.assertEqual(0, called)
        driver.advance(0.3)
        self.assertEqual(0, called)
        driver.advance(0.3)
        self.assertEqual(1, called)
        driver.advance(0.6)
        self.assertEqual(2, called)

    def test_canceling(self) -> None:
        """
        CallHandle.cancel() cancels an outstanding call.
        """
        scheduler = SimpleScheduler(driver := MemoryDriver())
        callTimes = []

        def record(event: str) -> Callable[[], None]:
            def result() -> None:
                callTimes.append((scheduler.now(), event))

            return result

        aHandle = scheduler.callAt(1.0, record("a"))
        bHandle = scheduler.callAt(2.0, record("b"))
        scheduler.callAt(3.0, record("c"))
        last = scheduler.callAt(2.5, record("d"))
        last.cancel()
        didCancel = []

        def bCancel() -> None:
            didCancel.append(True)
            bHandle.cancel()

        scheduler.callAt(1.5, bCancel)
        self.assertEqual(callTimes, [])
        driver.advance()
        self.assertEqual(callTimes, [(1.0, "a")])
        aHandle.cancel()  # if it's already called it's a no-op
        self.assertEqual(didCancel, [])
        driver.advance()
        self.assertEqual(callTimes, [(1.0, "a")])
        self.assertEqual(didCancel, [True])
        bHandle.cancel()  # repeated calls are no-ops
        self.assertEqual(didCancel, [True])
        driver.advance()
        self.assertEqual(callTimes, [(1.0, "a"), (3.0, "c")])

    def test_queueMustBeEmpty(self) -> None:
        driver = MemoryDriver()
        q = Heap([FutureCall(1.0, noop, 1, False, False, nocancel)])
        with self.assertRaises(ValueError):
            SimpleScheduler(driver, q)


def noop() -> None:
    ...


def nocancel(x: object) -> None:
    ...



# -*- test-case-name: fritter.test.test_tree -*-

"""
Groups of timers that may be paused, resumed, or time-scaled together.
"""

from __future__ import annotations
from dataclasses import dataclass
from .scheduler import FutureCall
from typing import Callable, NewType, Optional, Protocol, Tuple, TypeVar

from .scheduler import Scheduler

_BranchTime = NewType("_BranchTime", float)
_TrunkTime = NewType("_TrunkTime", float)


class Group(Protocol):
    """
    A L{Group} presents an interface to control a group of timers collected
    into a scheduler; pausing the group, unpausing it, or making its relative
    rate of progress faster or slower.
    """

    scaleFactor: float
    """
    How much faster the branch time coordinate system is within this scheduler?
    i.e.: with a scale factor of 2, that means time is running 2 times faster
    in this branch temporal coordinate system, and C{self.callAt(3.0,
    X)} will run C{X} when the trunk's current timestamp is 1.5.
    """

    def unpause(self) -> None:
        """
        Start the group of timers again.
        """

    def pause(self) -> None:
        """
        Pause the group of timers.
        """


def branch(
    trunk: Scheduler[float, Callable[[], None]], scaleFactor: float = 1.0
) -> tuple[Group, Scheduler[float, Callable[[], None]]]:
    """
    Derive a branch (child) scheduler from a trunk (trunk) scheduler.
    """
    driver = _BranchDriver(trunk)
    driver.scaleFactor = scaleFactor
    branchScheduler: Scheduler[float, Callable[[], None]] = Scheduler(driver)
    driver.unpause()
    return driver, branchScheduler


_F = TypeVar("_F", bound=float)


def _add(someFloat: _F, other: _F) -> _F:
    return someFloat + other  # type:ignore[return-value]


def _subtract(someFloat: _F, other: _F) -> _F:
    return someFloat - other  # type:ignore[return-value]


@dataclass
class _BranchDriver:
    """
    Implementation of L{TimeDriver} for L{Scheduler} that is stacked on top of
    another L{Scheduler}.
    """

    trunk: Scheduler[float, Callable[[], None]]
    """
    The scheduler that this driver is a branch of.
    """

    # TODO: support for generic types would be nice here, but:

    # 1. for the time type, things have to be multiplied and divided by the
    #    scaling factor, which means it needs to be a float. Even using a
    #    TypeVar bound to `float` here creates a ton of awkward casts below.

    # 2. for the work type, we have our own function that needs to be scheduled
    #    with the trunk scheduler, and so we couldn't meaningfully integrate
    #    with a higher-level persistent scheduler.

    _scaleFactor: float = 1.0
    """
    How much faster the branch time coordinate system is within this scheduler?
    i.e.: with a scale factor of 2, that means time is running 2 times faster
    in this branch temporal coordinate system, and C{self.callAt(3.0,
    X)} will run C{X} when the trunk's current timestamp is 1.5.
    """

    _call: Optional[FutureCall[_TrunkTime, Callable[[], None]]] = None

    _offset: _TrunkTime = _TrunkTime(0.0)
    """
    Amount to subtract from trunk's timestamp to get to this driver's base
    relative timestamp - in trunk's (unscaled, not branch) time-scale.  When a
    L{_BranchDriver} is created, it has a default offset of 0, which means that
    the moment '.start()' is called, that is time 0 in branch time.
    """

    _running: bool = False
    _pauseTime: _BranchTime = _BranchTime(0.0)
    """
    Timestamp at which we were last paused.
    """
    _fudge: _BranchTime = _BranchTime(0.0)

    _scheduleWhenStarted: Optional[
        Tuple[_BranchTime, Callable[[], None]]
    ] = None

    def _branchToTrunk(self, branchTime: _BranchTime) -> _TrunkTime:
        return _TrunkTime((branchTime / self._scaleFactor) + self._offset)

    def _trunkToBranch(self, trunkTime: _TrunkTime) -> _BranchTime:
        return _BranchTime((trunkTime - self._offset) * self._scaleFactor)

    @property
    def _trunk(self) -> Scheduler[_TrunkTime, Callable[[], None]]:
        return self.trunk  # type:ignore[return-value]

    def reschedule(self, desiredTime: float, work: Callable[[], None]) -> None:
        self._reschedule(_BranchTime(desiredTime), work)

    def _reschedule(
        self, desiredTime: _BranchTime, work: Callable[[], None]
    ) -> None:
        assert (
            self._call is None or self._running
        ), f"we weren't running, call should be None not {self._call}"
        self._scheduleWhenStarted = desiredTime, work
        if not self._running:
            return

        if self._call is not None:
            self._call.cancel()

        trunkTimestamp = self._branchToTrunk(desiredTime)
        roundTripped: _BranchTime = self._trunkToBranch(trunkTimestamp)
        self._fudge = _subtract(desiredTime, roundTripped)

        def clearAndRun() -> None:
            self._scheduleWhenStarted = None
            self._call = None
            work()

        self._call = self._trunk.callAt(trunkTimestamp, clearAndRun)

    def unschedule(self) -> None:
        self._scheduleWhenStarted = None
        if self._call is not None:
            self._call.cancel()
            self._call = None

    def now(self) -> float:
        return self._now()

    def _now(self) -> _BranchTime:
        if self._running:
            return _add(self._trunkToBranch(self._trunk.now()), self._fudge)
        else:
            return self._pauseTime

    # implementation of 'Group' interface

    def unpause(self) -> None:
        if self._running:
            return
        # shift forward the offset to skip over the time during which we were
        # paused.
        trunkTime: _TrunkTime = self._trunk.now()
        trunkDelta: _TrunkTime = _TrunkTime(
            self._pauseTime / self._scaleFactor
        )

        # We need to cast to the NewType again here because the results of
        # NewType arithmetic are the base types.
        self._offset = _TrunkTime(trunkTime - trunkDelta)
        self._pauseTime = _BranchTime(0.0)
        self._running = True
        scheduleWhenStarted = self._scheduleWhenStarted
        if scheduleWhenStarted is not None:
            desiredTime, work = scheduleWhenStarted
            self._reschedule(desiredTime, work)

    def pause(self) -> None:
        self._pauseTime = self._now()
        self._running = False
        if self._call is not None:
            self._call.cancel()
            self._call = None

    @property
    def scaleFactor(self) -> float:
        """
        The scale factor is how much faster than its trunk time passes in this
        driver.
        """
        return self._scaleFactor

    @scaleFactor.setter
    def scaleFactor(self, newScaleFactor: float) -> None:
        """
        Change this recursive driver to be running at `newScaleFactor` times
        its trunk scheduler's rate.  i.e. driver.changeScaleFactor(3.0) will
        change this driver's rate of time passing to be 3x faster than its
        trunk.
        """
        wasRunning = self._running
        if wasRunning:
            self.pause()
            self._scaleFactor = newScaleFactor
            self.unpause()
        else:
            self._scaleFactor = newScaleFactor



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from zoneinfo import ZoneInfo

from datetype import aware
from fritter.drivers.datetime import DateTimeDriver
from fritter.drivers.memory import MemoryDriver
from fritter.persistent.json import JSONableScheduler, JSONObject, JSONRegistry

registry = JSONRegistry[dict[str, str]]()

@dataclass
class MyClass:
    value: int

    @classmethod
    def typeCodeForJSON(cls) -> str:
        return ".".join([cls.__module__, cls.__name__])

    def asJSON(self) -> dict[str, object]:
        return {"value": self.value}

    @classmethod
    def fromJSON(
        cls,
        registry: JSONRegistry[dict[str, str]],
        scheduler: JSONableScheduler,
        loadContext: dict[str, str],
        json: JSONObject,
    ) -> MyClass:
        return cls(json["value"])

    @registry.method
    def later(self) -> None:
        print("my value is", self.value)


memoryDriver = MemoryDriver()
scheduler = JSONableScheduler(DateTimeDriver(memoryDriver))
dt = aware(
    datetime(
        2023,
        7,
        21,
        1,
        1,
        1,
        tzinfo=ZoneInfo(key="America/Los_Angeles"),
    ),
    ZoneInfo,
)

handle = scheduler.callAt(dt, MyClass(3).later)
myInstance = MyClass(3)
from json import dumps, loads
dump = dumps(registry.save(scheduler))
print(dump)
mem2 = MemoryDriver()
loaded = registry.load(mem2, loads(dump), {})
mem2.advance(dt.timestamp())



from fritter.boundaries import Cancellable
from fritter.repeat import Async, EverySecond
from fritter.drivers.asyncio import AsyncioTimeDriver, AsyncioAsyncDriver
from fritter.scheduler import SimpleScheduler
from asyncio import run

# example coroutine
async def example() -> None:
    scheduler = SimpleScheduler(AsyncioTimeDriver())
    repeatedly = Async(AsyncioAsyncDriver()).repeatedly
    times = 0

    # work function
    async def work(steps: int, stopper: Cancellable) -> None:
        nonlocal times
        times += steps
        print(times, f"times: {times} {scheduler.now()}", flush=True)
        if times > 3:
            stopper.cancel()

    await repeatedly(scheduler, EverySecond(0.25), work)

# run
run(example())



# setup
from fritter.drivers.memory import MemoryDriver
from fritter.scheduler import SimpleScheduler
from fritter.tree import branch

driver = MemoryDriver()
parent = SimpleScheduler(driver)
group, child = branch(parent, 3.0)
# end setup


# showfunc
def loop(scheduler: SimpleScheduler, name: str, interval: float = 1.0) -> None:
    def _() -> None:
        print(name)
        scheduler.callAt(scheduler.now() + interval, _)

    _()


# end showfunc

# loops
loop(parent, "parent", 1.0)
loop(child, "child", 1.0)
# work
for again in range(10):
    driver.advance()
    group.scaleFactor += 1.0
    print(f"time: parent={parent.now()} child={child.now()}")



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import Any
from unittest import TestCase
from zoneinfo import ZoneInfo
from json import dumps, loads

from datetype import aware

from ..boundaries import TimeDriver, Cancellable
from ..drivers.datetime import DateTimeDriver
from ..drivers.memory import MemoryDriver
from ..persistent.json import JSONableScheduler, JSONObject, JSONRegistry
from ..repeat import daily
from ..scheduler import Scheduler


@dataclass
class RegInfo:
    calls: list[str]
    identityMap: dict[str, Any] = field(default_factory=dict)


registry = JSONRegistry[RegInfo]()
emptyRegistry = JSONRegistry[RegInfo]()

calls = []


@registry.function
def call1() -> None:
    calls.append("hello")


@registry.function
def call2() -> None:
    calls.append("goodbye")


@registry.repeatFunction
def repeatable(steps: int, stopper: Cancellable) -> None:
    calls.append(f"repeatable {steps}")


@dataclass
class InstanceWithMethods:
    value: str
    info: RegInfo
    calls: int = 0

    @classmethod
    def typeCodeForJSON(self) -> str:
        return "instanceWithMethods"

    @classmethod
    def fromJSON(
        cls,
        registry: JSONRegistry[RegInfo],
        scheduler: JSONableScheduler,
        loadContext: RegInfo,
        json: JSONObject,
    ) -> InstanceWithMethods:
        key = json["identity"]
        if key in loadContext.identityMap:
            loadContext.calls.append("InstanceWithMethods.fromJSON (cached)")
            self: InstanceWithMethods = loadContext.identityMap[key]
            return self
        loadContext.calls.append("InstanceWithMethods.fromJSON")
        new = cls(json["value"], loadContext)
        loadContext.identityMap[key] = new
        return new

    def asJSON(self) -> dict[str, object]:
        return {
            "value": self.value,
            "identity": id(self),
        }

    @registry.method
    def method1(self) -> None:
        self.info.calls.append(f"{self.value}/method1")

    @registry.method
    def method2(self) -> None:
        self.info.calls.append(f"{self.value}/method2")

    @registry.repeatMethod
    def repeatMethod(self, steps: int, stopper: Cancellable) -> None:
        self.calls += 1
        self.info.calls.append(
            f"repeatMethod {steps} {self.value=} {self.calls=}"
        )


def jsonScheduler(driver: TimeDriver[float]) -> JSONableScheduler:
    return Scheduler(DateTimeDriver(driver))


class PersistentSchedulerTests(TestCase):
    def tearDown(self) -> None:
        del calls[:]

    def test_scheduleRunSaveRun(self) -> None:
        """
        Test scheduling module-level functions and instance methods.
        """
        memoryDriver = MemoryDriver()
        scheduler = jsonScheduler(memoryDriver)
        dt = aware(
            datetime(
                2023,
                7,
                21,
                1,
                1,
                1,
                tzinfo=ZoneInfo(key="America/Los_Angeles"),
            ),
            ZoneInfo,
        )
        dt2 = aware(
            datetime(
                2023,
                7,
                22,
                1,
                1,
                1,
                tzinfo=ZoneInfo(key="America/Los_Angeles"),
            ),
            ZoneInfo,
        )
        ri0 = RegInfo([])
        iwm = InstanceWithMethods("test_scheduleRunSaveRun value", ri0)
        scheduler.callAt(dt, call1)
        scheduler.callAt(dt, iwm.method1)
        scheduler.callAt(dt2, call2)
        scheduler.callAt(dt2, iwm.method2)
        memoryDriver.advance(dt.timestamp() + 1)
        self.assertEqual(calls, ["hello"])
        del calls[:]
        saved = registry.save(scheduler)
        memory2 = MemoryDriver()
        ri = RegInfo([])
        registry.load(memory2, saved, ri)
        memory2.advance(dt2.timestamp() + 1)
        self.assertEqual(calls, ["goodbye"])
        self.assertEqual(ri0.calls, ["test_scheduleRunSaveRun value/method1"])
        self.assertEqual(
            ri.calls,
            [
                "InstanceWithMethods.fromJSON",
                "test_scheduleRunSaveRun value/method2",
            ],
        )

    def test_idling(self) -> None:
        memoryDriver = MemoryDriver()
        scheduler = jsonScheduler(memoryDriver)
        dt = aware(
            datetime(
                2023,
                7,
                21,
                1,
                1,
                1,
                tzinfo=ZoneInfo(key="America/Los_Angeles"),
            ),
            ZoneInfo,
        )
        handle = scheduler.callAt(dt, call1)
        self.assertEqual(memoryDriver.isScheduled(), True)
        handle.cancel()
        self.assertEqual(memoryDriver.isScheduled(), False)
        memoryDriver.advance(dt.timestamp() + 1)
        self.assertEqual(calls, [])

    def test_emptyScheduler(self) -> None:
        memory = MemoryDriver()
        registry.load(memory, {"scheduledCalls": []}, RegInfo([]))
        self.assertEqual(memory.isScheduled(), False)

    def test_repeatable(self) -> None:
        dt = aware(
            datetime(
                2023,
                7,
                21,
                1,
                1,
                1,
                tzinfo=ZoneInfo(key="America/Los_Angeles"),
            ),
            ZoneInfo,
        )
        memoryDriver = MemoryDriver()
        memoryDriver.advance(dt.timestamp())
        scheduler = jsonScheduler(memoryDriver)
        registry.repeatedly(scheduler, daily, repeatable, dt)
        self.assertEqual(calls, ["repeatable 1"])
        del calls[:]

        def days(n: int) -> float:
            return 60 * 60 * 24 * n

        memoryDriver.advance(days(3))
        self.assertEqual(calls, ["repeatable 3"])
        del calls[:]

        newInfo = RegInfo([])
        mem2 = MemoryDriver()
        mem2.advance(dt.timestamp())
        mem2.advance(days(7))
        self.assertEqual(mem2.isScheduled(), False)
        registry.load(mem2, loads(dumps(registry.save(scheduler))), newInfo)
        self.assertEqual(mem2.isScheduled(), True)
        amount = mem2.advance()
        assert amount is not None
        self.assertLess(amount, 0.0001)
        self.assertEqual(calls, ["repeatable 4"])

    def test_repeatableMethod(self) -> None:
        dt = aware(
            datetime(
                2023,
                7,
                21,
                1,
                1,
                1,
                tzinfo=ZoneInfo(key="America/Los_Angeles"),
            ),
            ZoneInfo,
        )
        memoryDriver = MemoryDriver()
        memoryDriver.advance(dt.timestamp())
        scheduler = jsonScheduler(memoryDriver)
        info = RegInfo([])
        method = InstanceWithMethods("sample", info).repeatMethod
        shared = InstanceWithMethods("shared", info)
        registry.repeatedly(scheduler, daily, method)
        registry.repeatedly(scheduler, daily, shared.repeatMethod)
        registry.repeatedly(scheduler, daily, shared.repeatMethod)
        self.assertEqual(
            info.calls,
            [
                "repeatMethod 1 self.value='sample' self.calls=1",
                "repeatMethod 1 self.value='shared' self.calls=1",
                "repeatMethod 1 self.value='shared' self.calls=2",
            ],
        )
        del info.calls[:]

        def days(n: int) -> float:
            return 60 * 60 * 24 * n

        memoryDriver.advance(days(3))
        self.assertEqual(
            info.calls,
            [
                "repeatMethod 3 self.value='sample' self.calls=2",
                "repeatMethod 3 self.value='shared' self.calls=3",
                "repeatMethod 3 self.value='shared' self.calls=4",
            ],
        )

        newInfo = RegInfo([])
        mem2 = MemoryDriver()
        mem2.advance(dt.timestamp())
        mem2.advance(days(7))
        self.assertEqual(mem2.isScheduled(), False)
        persistent = dumps(registry.save(scheduler))
        registry.load(mem2, loads(persistent), newInfo)
        loaded = loads(persistent)
        with self.assertRaises(KeyError):
            # TODO: allow for better error handling that doesn't just blow up
            emptyRegistry.load(MemoryDriver(), loaded, newInfo)
        self.assertEqual(mem2.isScheduled(), True)
        mem2.advance()
        self.assertEqual(
            newInfo.calls,
            [
                "InstanceWithMethods.fromJSON",
                "InstanceWithMethods.fromJSON",
                "InstanceWithMethods.fromJSON (cached)",
                "repeatMethod 4 self.value='sample' self.calls=1",
                "repeatMethod 4 self.value='shared' self.calls=1",
                "repeatMethod 4 self.value='shared' self.calls=2",
            ],
        )



from typing import Any, Callable
from unittest import TestCase

from twisted.internet.defer import CancelledError, Deferred, succeed

from ..boundaries import Cancellable
from ..drivers.memory import MemoryDriver
from ..drivers.twisted import TwistedAsyncDriver
from ..repeat import EverySecond, Async, repeatedly
from ..scheduler import Scheduler


class RepeatTestCase(TestCase):
    def test_synchronous(self) -> None:
        mem = MemoryDriver()
        calls = []

        def work(steps: int, stopper: Cancellable) -> None:
            now = mem.now()
            if mem.now() >= 10.0:
                stopper.cancel()
            calls.append((steps, now))

        repeatedly(Scheduler(mem), work, EverySecond(5))

        self.assertTrue(mem.isScheduled())
        self.assertEqual(calls, [(1, 0.0)])
        calls = []
        mem.advance()
        self.assertTrue(mem.isScheduled())
        self.assertEqual(calls, [(1, 5.0)])
        calls = []
        mem.advance()
        self.assertFalse(mem.isScheduled())
        self.assertEqual(calls, [(1, 10.0)])

    def test_repeatEveryIntervalInSeconds(self) -> None:
        tad = TwistedAsyncDriver()
        mem = MemoryDriver()
        count = 0
        calls = []
        event: Deferred[None] = succeed(None)

        async def tick(times: int) -> None:
            nonlocal count, event
            calls.append(f"before {count} ({times})")
            event = Deferred()
            await event
            calls.append(f"after {count}")
            count += 1

        Async(tad).repeatedly(
            Scheduler(mem),
            EverySecond(15),
            lambda times, stopper: tick(times),
        )

        self.assertEqual(calls, ["before 0 (1)"])
        event.callback(None)
        self.assertEqual(calls, ["before 0 (1)", "after 0"])
        del calls[:]
        mem.advance(15.3)
        self.assertEqual(calls, ["before 1 (1)"])
        del calls[:]
        # async operation takes 45 seconds. it's now 60.3.
        mem.advance(45.0)
        self.assertEqual(calls, [])
        event.callback(None)
        # catch-up call is immediately scheduled
        self.assertEqual(calls, ["after 1", "before 2 (3)"])
        del calls[:]
        event.callback(None)
        self.assertEqual(calls, ["after 2"])

    def test_complete(self) -> None:
        count = 0
        threshold = 3

        def step(steps: int, stopper: Cancellable) -> Deferred[None]:
            nonlocal count
            count += steps
            if count >= threshold:
                stopper.cancel()
            return succeed(None)

        tad = TwistedAsyncDriver()
        mem = MemoryDriver()
        done = False

        async def task() -> None:
            nonlocal done
            await Async(tad).repeatedly(Scheduler(mem), EverySecond(1), step)
            done = True

        tad.runAsync(task())
        for ignored in range(threshold):
            mem.advance()
        self.assertTrue(done)
        self.assertEqual(count, threshold)

    def test_cancel(self) -> None:
        tad = TwistedAsyncDriver()
        mem = MemoryDriver()

        def canceled(d: Deferred[None]) -> None:
            return

        pending: Deferred[None] = Deferred(canceled)

        async def asynchronously() -> None:
            with self.assertRaises(CancelledError):
                await pending

        async def synchronously() -> None:
            pass

        repeatCall: Deferred[None]

        def go(how: Callable[[], Any]) -> None:
            nonlocal repeatCall
            repeatCall = Async(tad).repeatedly(
                Scheduler(mem),
                EverySecond(1),
                lambda times, stopper: how(),
            )

        async def run(how: Callable[[], Any]) -> None:
            go(how)
            with self.assertRaises(CancelledError):
                await repeatCall

        tad.runAsync(run(asynchronously))
        self.assertTrue(mem.isScheduled())
        repeatCall.cancel()
        self.assertFalse(mem.isScheduled())

        tad.runAsync(run(synchronously))
        self.assertTrue(mem.isScheduled())
        repeatCall.cancel()
        self.assertFalse(mem.isScheduled())



# setup
from typing import Callable
from fritter.drivers.memory import MemoryDriver
from fritter.scheduler import SimpleScheduler
from fritter.tree import branch

driver = MemoryDriver()
parent = SimpleScheduler(driver)
group, child = branch(parent)
# end setup


# showfunc
def show(name: str) -> Callable[[], None]:
    def _() -> None:
        print(f"{name} parent={parent.now()} child={child.now()}")

    return _


# end showfunc


# childcalls
child.callAt(1.0, show("child 1"))
child.callAt(2.0, show("child 2"))
child.callAt(3.0, show("child 3"))
# parentcalls
parent.callAt(1.0, show("parent 1"))
parent.callAt(2.0, show("parent 2"))
parent.callAt(3.0, show("parent 3"))
# endcalls

# interact
driver.advance()
print("pause")
group.pause()
driver.advance()
print("unpause")
group.unpause()
driver.advance()
driver.advance()



# -*- test-case-name: fritter.test.test_repeat -*-
"""
Schedule repeated invocations of a function, indicating how many steps have
been passed so that the repeated calls may catch up to real time to preserve
timing accuracy when timers cannot always be invoked promptly.
"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from typing import Any, Callable, Coroutine, Generic, Protocol, TypeVar
from zoneinfo import ZoneInfo

from datetype import DateTime

from fritter.boundaries import Cancellable

from .boundaries import AsyncDriver, AsyncType, RepeatingWork
from .scheduler import Scheduler, WhatT, WhenT


class RecurrenceRule(Protocol[WhenT]):
    """
    A L{RecurrenceRule} is a callable that takes a reference time and a current
    time, and computes a step count for the current recurrence and a new
    reference time for the next one.
    """

    def __call__(self, reference: WhenT, current: WhenT) -> tuple[int, WhenT]:
        """
        Given a reference time and a current time, compute a step count and the
        next reference time.

        @param reference: (the time at which the current recurrence was
            scheduled to occur)

        @return: a 2-tuple of:

                1. a I{step count}; the number of recurrences that were
                   expected to have occurred between C{reference} and the
                   I{current time}.  So for example, for a L{RecurrenceRule}
                   representing a once-every-5-seconds recurrence, if your
                   reference time were 1.0 and your current time were 15.0,
                   then your step count should be 2, since recurrences should
                   have occurred at 6.0 and 11.0.

                2. a I{desired time}; the next reference time at which a
                   recurrence should occur.  In our previous example, where our
                   reference time was 1.0 and current time was 15.0, the next
                   desired time should be 16.0, since that's the next 5-second
                   recurrence after 11.0.
        """


RepeatingWhatT = TypeVar("RepeatingWhatT", bound=RepeatingWork)
"""
A TypeVar for L{Repeater} to reference a specific type of L{RepeatingWork}.
"""

DTRule = RecurrenceRule[DateTime[ZoneInfo]]
"""
A type alias to describe a recurrence rule function that operates on aware
datetimes.
"""


@dataclass(frozen=True)
class EveryDelta:
    """
    An L{EveryDelta} is a L{RecurrenceRule} based on a L{timedelta}, that can
    cause civil times to repeat.

    @ivar delta: the time delta between recurrences
    """

    delta: timedelta

    def __call__(
        self,
        reference: DateTime[ZoneInfo],
        current: DateTime[ZoneInfo],
    ) -> tuple[int, DateTime[ZoneInfo]]:
        """
        Compute a step count and next desired recurrence time based on a
        reference time and a current time in aware datetimes, and the timedelta
        in C{self.delta}.

        @see: L{RecurrenceRule}
        """
        count = 0
        nextDesired = reference
        while nextDesired <= current:
            count += 1
            nextDesired += self.delta
        return count, nextDesired


@dataclass
class EverySecond:
    """
    An L{EveryDelta} is a L{RecurrenceRule} based on a L{timedelta}, that can
    cause civil times to repeat.

    @ivar seconds: the number of seconds between recurrences.
    """

    seconds: float

    def __call__(self, reference: float, current: float) -> tuple[int, float]:
        """
        Compute a step count and next desired recurrence time based on a
        reference time and a current time in floating-point POSIX timestamps
        like those returned by L{time.time} and L{datetime.datetime.timestamp},
        and number of seconds in L{self.seconds <EverySecond.seconds>}.

        @see: L{RecurrenceRule}
        """
        elapsed = current - reference
        count, remainder = divmod(elapsed, self.seconds)
        return int(count) + 1, current + (self.seconds - remainder)


weekly: DTRule = EveryDelta(timedelta(weeks=1))
"""
Weekly datetime-based delta.
"""

daily: DTRule = EveryDelta(timedelta(days=1))
"""
Daily datetime-based rule.
"""
hourly: DTRule = EveryDelta(timedelta(hours=1))
"""
Hourly datetime-based rule.
"""

_everyIsRecurrenceRule: Callable[[float], RecurrenceRule[float]]
_everyIsRecurrenceRule = EverySecond


@dataclass
class Repeater(Generic[WhenT, WhatT, RepeatingWhatT]):
    """
    A L{Repeater} can call a L{RepeatingWork} function repeatedly.

    @ivar scheduler: The scheduler where the work will be performed.

    @ivar rule: The L{RecurrenceRule} that defines the times at which C{work}
        will be invoked.

    @ivar work: The L{RepeatingWork} that will be performed.

    @ivar convert: Since a L{Scheduler} requires that any work provided be
        specifically of its C{WhatT} type, which may need to have some property
        associated with it beyond its function signature (for example,
        additional attributes for instrumentation or serialization), so this
        C{convert} function will convert this L{Repeater} into a 0-argument
        function that both:

            1. is of the appropriate L{WhatT} for C{scheduler}, and

            2. invokes this L{Repeater}'s C{repeat} method.

    @ivar reference: The current reference time, i.e. the time at which the
        next invocation of C{work} I{should} occur.
    """

    scheduler: Scheduler[WhenT, WhatT]
    rule: RecurrenceRule[WhenT]
    work: RepeatingWhatT
    convert: Callable[[Repeater[WhenT, WhatT, RepeatingWhatT]], WhatT]
    reference: WhenT

    @classmethod
    def new(
        cls,
        scheduler: Scheduler[WhenT, WhatT],
        rule: RecurrenceRule[WhenT],
        work: RepeatingWhatT,
        convert: Callable[[Repeater[WhenT, WhatT, RepeatingWhatT]], WhatT],
        reference: WhenT | None = None,
    ) -> Repeater[WhenT, WhatT, RepeatingWhatT]:
        """
        Create a L{Repeater}, filling out its reference time with the L{current
        time of the given scheduler <Scheduler.now>}, if no other time is
        provided.
        """
        if reference is None:
            reference = scheduler.now()
        return cls(scheduler, rule, work, convert, reference)

    def repeat(self) -> None:
        """
        Repeat the L{work <Repeater.work>} associated with this L{Repeater}.

        Applications should call this once, and only once, after the
        L{Repeater} is created, in order to kick off the repetition.  All
        future calls should be performed via the result of C{convert} being
        called in the scheduler, or the timing of repeated invocations will be
        incorrect.
        """
        now = self.scheduler.now()
        callIncrement, self.reference = self.rule(self.reference, now)
        callRepeat = self.convert(self)
        stopHandle = self.scheduler.callAt(self.reference, callRepeat)
        self.work(callIncrement, stopHandle)


def repeatedly(
    scheduler: Scheduler[WhenT, Callable[[], None]],
    work: RepeatingWork,
    rule: RecurrenceRule[WhenT],
) -> None:
    """
    Create a L{Repeater} and call its C{repeat} method.  This is a utility
    function for use when you have a simple repetition set up on a scheduler
    that accepts a baseline 0-argument callable, and does not require any
    L{conversion <Repeater.convert>}.
    """
    Repeater.new(scheduler, rule, work, lambda r: r.repeat).repeat()


@dataclass
class _AsyncStopper(Generic[AsyncType]):
    """
    An implementation of L{Cancellable} which can stop the repetition kicked
    off by L{Async.repeatedly}.
    """

    driver: AsyncDriver[AsyncType]
    result: AsyncType
    timeInProgress: Cancellable | None = None
    asyncInProgress: Cancellable | None = None
    shouldComplete: bool = True

    def cancel(self) -> None:
        if self.timeInProgress is not None:
            self.timeInProgress.cancel()
        if self.asyncInProgress is not None:
            self.asyncInProgress.cancel()
        if self.shouldComplete:
            self.driver.complete(self.result)


@dataclass
class Async(Generic[AsyncType]):
    """
    An L{Async} wraps an L{AsyncDriver} and provides an implementation of
    L{repeatedly} which:

        1. C{await}s each result from its async C{work} callable, so that no
           overlapping work will be performed if an asynchronous operation
           takes longer than the repetition interval,

        2. returns an awaitable that fires when C{stopper.cancel()} has been
           called on the C{stopper} provided to that callable, and

        3. provides a C{.cancel()} implementation on that returned awaitable
           which stops any in-progress async work and raises the appropriate
           cancellation error for your framework back to the caller.

    @ivar asyncDriver: The driver that supplies awaitables for this L{Async} to
        return.
    """

    asyncDriver: AsyncDriver[AsyncType]

    def repeatedly(
        self,
        scheduler: Scheduler[WhenT, Callable[[], None]],
        rule: RecurrenceRule[WhenT],
        work: Callable[
            [int, Cancellable], AsyncType | Coroutine[AsyncType, Any, Any]
        ],
    ) -> AsyncType:
        """
        Kick off a repeated call within the given scheduler, returning an
        L{AsyncType} (i.e. L{Future <asyncio.Future>}, L{Deferred
        <twisted.internet.defer.Deferred>}, or similar, as defined by
        L{Async.asyncDriver}).

        If you call C{.cancel()} on the result of this method, the repetition
        will be stopped and a cancellation error will be signaled to the
        caller.  If, instead, C{work} calls its C{.cancel()} method, then the
        result of this method will complete successfully, returning C{None} to
        a coroutine awaiting it.
        """

        def reallyCancel() -> None:
            asyncStopper.shouldComplete = False
            asyncStopper.cancel()

        asyncStopper: _AsyncStopper[AsyncType] = _AsyncStopper(
            self.asyncDriver,
            self.asyncDriver.newWithCancel(reallyCancel),
        )

        def complete() -> None:
            asyncStopper.asyncInProgress = None
            if asyncStopper.timeInProgress is None:
                repeater.repeat()

        def kickoff(steps: int, stopper: Cancellable) -> None:
            asyncStopper.timeInProgress = stopper

            if asyncStopper.asyncInProgress is not None:
                return

            completedSynchronously: bool = False

            async def coro() -> None:
                nonlocal completedSynchronously
                try:
                    await work(steps, asyncStopper)
                finally:
                    if asyncStopper.asyncInProgress is None:
                        completedSynchronously = True
                    else:
                        complete()

            asyncStopper.asyncInProgress = self.asyncDriver.runAsync(coro())
            if completedSynchronously:
                complete()

        def whenReady() -> None:
            asyncStopper.timeInProgress = None
            if asyncStopper.asyncInProgress is None:
                repeater.repeat()

        repeater = Repeater.new(scheduler, rule, kickoff, lambda r: whenReady)
        repeater.repeat()

        return asyncStopper.result



# -*- test-case-name: fritter.test.test_sleep -*-
"""
Implementation of L{TimeDriver} that can run timers by blocking until all its
timers have been run, sleeping between them as necessary.

This is suitable for batch scripts that don't require an event loop like
L{asyncio <fritter.drivers.asyncio>} or L{twisted <fritter.drivers.twisted>}.
"""
from dataclasses import dataclass
from time import sleep as _sleep, time as _time
from typing import Callable, TYPE_CHECKING
from math import inf

from ..boundaries import TimeDriver


@dataclass
class SleepDriver:
    """
    Instantiate a L{SleepDriver} with no arguments to get one that sleeps using
    L{time.sleep} and gets the current time with L{time.time}.

    For testing, you can supply those parameters, but for most test cases, you
    should probably prefer a L{fritter.drivers.memory.MemoryDriver}.

    @ivar sleep: The L{time.sleep}-like callable that this driver will use to
        sleep for a given number of seconds.

    @ivar time: The L{time.time}-like callable that this driver will use to get
        the current time.
    """

    sleep: Callable[[float], None] = _sleep
    time: Callable[[], float] = _time
    _work: tuple[float, Callable[[], None]] | None = None

    def reschedule(self, desiredTime: float, work: Callable[[], None]) -> None:
        "Implementation of L{TimeDriver.reschedule}"
        self._work = desiredTime, work

    def unschedule(self) -> None:
        "Implementation of L{TimeDriver.unschedule}"
        self._work = None

    def now(self) -> float:
        "Implementation of L{TimeDriver.now}"
        return self.time()

    def block(self, timeout: float = inf) -> int:
        """
        While any active timer is scheduled with L{reschedule
        <SleepDriver.reschedule>}, sleep until the desired time specified by
        that call, call the work that was scheduled, and repeat.

        @param timeout: if specified, the maximum I{total} amount of time to
            sleep before unblocking, even if work remains to be done.
        """
        worked = 0
        maxTime = self.time() + timeout
        while True:
            scheduled = self._work
            if scheduled is None:
                break
            time, work = scheduled
            now = self.time()
            self._work, (time, work) = None, scheduled
            self.sleep(max(0, min(time, maxTime) - now))
            if time > maxTime:
                break
            worked += 1
            work()
        return worked


if TYPE_CHECKING:
    _CheckSleepDriver: type[TimeDriver[float]] = SleepDriver



from __future__ import annotations

from twisted.internet.task import Clock
from twisted.trial.unittest import SynchronousTestCase

from ..drivers.twisted import TwistedAsyncDriver, TwistedTimeDriver, scheduler


class TestAsyncDriver(SynchronousTestCase):
    def setUp(self) -> None:
        self.calls = 0

    def called(self) -> None:
        self.calls += 1

    def test_complete(self) -> None:
        driver = TwistedAsyncDriver()
        d = driver.newWithCancel(self.called)
        self.assertNoResult(d)
        driver.complete(d)
        self.assertIsNone(self.successResultOf(d))

    def test_cancel(self) -> None:
        driver = TwistedAsyncDriver()
        d = driver.newWithCancel(self.called)
        self.assertNoResult(d)
        d.cancel()
        self.failureResultOf(d)

    def test_runAsync(self) -> None:
        driver = TwistedAsyncDriver()
        operation = driver.newWithCancel(self.called)

        before = 0
        after = 0

        async def aCoroutine() -> None:
            nonlocal before, after
            before += 1
            await operation
            after += 1

        self.assertEqual(before, 0)
        self.assertEqual(after, 0)
        driver.runAsync(aCoroutine())
        self.assertEqual(before, 1)
        self.assertEqual(after, 0)
        driver.complete(operation)
        self.assertEqual(after, 1)


class TestTimeDriver(SynchronousTestCase):
    def setUp(self) -> None:
        self.calls = 0

    def called(self) -> None:
        self.calls += 1

    def test_now(self) -> None:
        clock = Clock()
        driver = TwistedTimeDriver(clock)
        clock.advance(1234)
        self.assertEqual(driver.now(), 1234.0)
        self.assertEqual(clock.getDelayedCalls(), [])

    def test_schedule(self) -> None:
        clock = Clock()
        driver = TwistedTimeDriver(clock)
        driver.reschedule(1.0, self.called)
        clock.advance(0.5)
        self.assertEqual(self.calls, 0)
        clock.advance(0.6)
        self.assertEqual(self.calls, 1)
        self.assertEqual(clock.getDelayedCalls(), [])

    def test_reschedule(self) -> None:
        clock = Clock()
        driver = TwistedTimeDriver(clock)
        driver.reschedule(1.0, self.called)
        clock.advance(0.5)
        self.assertEqual(self.calls, 0)
        driver.reschedule(2.0, self.called)
        clock.advance(0.6)
        self.assertEqual(self.calls, 0)
        clock.advance(0.9)
        self.assertEqual(self.calls, 1)
        self.assertEqual(clock.getDelayedCalls(), [])

    def test_unschedule(self) -> None:
        clock = Clock()
        driver = TwistedTimeDriver(clock)
        driver.reschedule(1.0, self.called)
        clock.advance(0.5)
        self.assertEqual(self.calls, 0)
        driver.unschedule()
        self.assertEqual(self.calls, 0)
        clock.advance(0.9)
        self.assertEqual(self.calls, 0)
        driver.unschedule()  # no-op, no exception
        self.assertEqual(clock.getDelayedCalls(), [])

    def test_schedulerDefault(self) -> None:
        sched = scheduler()
        self.assertIsInstance(sched.driver, TwistedTimeDriver)

    def test_scheduler(self) -> None:
        sched = scheduler(clock := Clock())
        stuff = []

        def hello() -> None:
            stuff.append("hello")

        sched.callAt(50, hello)
        self.assertEqual(stuff, [])
        clock.advance(60)
        self.assertEqual(stuff, ["hello"])
